@@ -2,7 +2,7 @@
 # -*- perl -*-
 
 #
-# $Id: FontDialog.pm,v 1.11 1999/06/09 23:45:19 eserte Exp $
+# $Id: FontDialog.pm,v 1.17 2002/10/23 09:11:18 eserte Exp $
 # Author: Slaven Rezic
 #
 # Copyright (C) 1998,1999 Slaven Rezic. All rights reserved.
@@ -23,15 +23,14 @@ use vars qw($VERSION @ISA);
 
 Construct Tk::Widget 'FontDialog';
 
-$VERSION = '0.04';
+$VERSION = '0.07';
 
 sub Populate {
     my($w, $args) = @_;
 
     require Tk::HList;
     require Tk::ItemStyle;
-    require Tk::Adjuster;
-    
+
     $w->SUPER::Populate($args);
     $w->protocol('WM_DELETE_WINDOW' => ['Cancel', $w ]);
 
@@ -40,8 +39,15 @@ sub Populate {
     if (exists $args->{-font}) {
 	$w->optionAdd('*font' => delete $args->{-font});
     }
-    my $dialog_font = $w->fontCreate($w->fontActual
-				     ($w->optionGet("font", "*")));
+    my $dialog_font;
+    my $font_name = $w->optionGet("font", "*");
+    if (!defined $font_name) {
+	my $l = $w->Label;
+	$dialog_font = $l->cget(-font);
+	warn $dialog_font;
+    } else {
+	$dialog_font = $w->fontCreate($w->fontActual($font_name));
+    }
     if (exists $args->{-initfont}) {
 	$w->{'curr_font'} = $w->fontCreate($w->fontActual
 					   (delete $args->{-initfont}));
@@ -58,9 +64,8 @@ sub Populate {
     my $overstrike_font = $w->fontCreate($w->fontActual($dialog_font),
 					 -overstrike => 1);
 
-    my $f1     = $w->Frame->pack#Adjust XXX
-      (-expand => 1, -fill => 'both',
-				       -padx => 2, -pady => 2);
+    my $f1     = $w->Frame->pack(-expand => 1, -fill => 'both',
+				 -padx => 2, -pady => 2);
     my $ffam   = $f1->Frame->pack(-expand => 1, -fill => 'both',
 				  -side => 'left');
     my $fsize  = $f1->Frame->pack(-expand => 1, -fill => 'both',
@@ -68,7 +73,7 @@ sub Populate {
     my $fstyle = $f1->Frame->pack(-expand => 1, -fill => 'both',
 				  -side => 'left');
 
-    my(%family_res) = _get_label(delete $args->{'-familylabel'} 
+    my(%family_res) = _get_label(delete $args->{'-familylabel'}
 				 || '~Family:');
     $ffam->Label
       (@{$family_res{'args'}},
@@ -81,7 +86,7 @@ sub Populate {
        -selectmode => 'single',
        -bg => 'white',
        -browsecmd => sub {
-	   my $family = $w->{'familiy_index'}[$_[0]];
+	   my $family = $w->{'family_index'}[$_[0]];
 	   $w->UpdateFont(-family => $family)
        },
       )->pack(-expand => 1, -fill => 'both', -anchor => 'w');
@@ -103,6 +108,7 @@ sub Populate {
        -browsecmd => sub { $w->UpdateFont(-size => $_[0]) },
       )->pack(-expand => 1, -fill => 'both', -anchor => 'w');
     $w->Advertise('size_list' => $sizelb);
+    $sizelb->bind("<3>" => [ $w, '_custom_size' ]);
 
     my @fontsizes;
     if (exists $args->{-fontsizes}) {
@@ -116,6 +122,7 @@ sub Populate {
 	$sizelb->add($size, -text => $size);
 	if ($size == $curr_size) {
 	    $sizelb->selectionSet($size);
+	    $sizelb->anchorSet($size);
 	    $sizelb->see($size);
 	}
     }
@@ -190,7 +197,7 @@ sub Populate {
        -fg => 'green4',
        -font => $bold_font,
        -command => ['Accept', $w ],
-      )->grid(-column => 0, -row => 0,
+      )->grid(-column => 0, -row => 0, -rowspan => 2,
 	      -sticky => 'ew', -padx => 5);
 
     my(%apply_res) = _get_label(delete $args->{'-applylabel'}
@@ -204,7 +211,7 @@ sub Populate {
 	   -fg => 'yellow4',
 	   -font => $bold_font,
 	   -command => sub { $applycmd->($w->ReturnFont($w->{'curr_font'})) },
-	  )->grid(-column => 1, -row => 0,
+	  )->grid(-column => 1, -row => 0, -rowspan => 2,
 		  -sticky => 'ew', -padx => 5);
     }
 
@@ -215,7 +222,7 @@ sub Populate {
        -fg => 'red',
        -font => $bold_font,
        -command => ['Cancel', $w ],
-      )->grid(-column => 2, -row => 0,
+      )->grid(-column => 2, -row => 0, -rowspan => 2,
 	      -sticky => 'ew', -padx => 5);
     $bf->grid('columnconfigure', 3, -weight => 1.0);
 
@@ -226,17 +233,34 @@ sub Populate {
        -variable => \$w->{'alt_sample'},
        -command => sub { $w->UpdateFont; },
       )->grid(-column => 4, -row => 0,
-	      -sticky => 'ew', -padx => 5);
+	      -sticky => 'w', -padx => 5);
 
-    my(%nicefonts_res) = _get_label(delete $args->{'-nicefontslabel'}
+    my(%nicefonts_res, $nicecb);
+    if (!exists $args->{'-nicefontsbutton'} || $args->{'-nicefontsbutton'}) {
+	%nicefonts_res = _get_label(delete $args->{'-nicefontslabel'}
 				    || "~Nicefonts");
-    my $nicecb = $bf->Checkbutton
-      (@{$nicefonts_res{'args'}},
-       -variable => \$w->{Configure}{-nicefont},
-       -command => sub { $w->InsertFamilies; },
-      )->grid(-column => 5, -row => 0,
-	      -sticky => 'ew', -padx => 5);
-    
+	$nicecb = $bf->Checkbutton
+	  (@{$nicefonts_res{'args'}},
+	   -variable => \$w->{Configure}{-nicefont},
+	   -command => sub { $w->InsertFamilies; },
+	  )->grid(-column => 4, -row => 1,
+		  -sticky => 'w', -padx => 5);
+    }
+    delete $args->{'-nicefontsbutton'};
+
+    my(%fixedfonts_res, $fixedcb);
+    if (!exists $args->{'-fixedfontsbutton'} || $args->{'-fixedfontsbutton'}) {
+	%fixedfonts_res = _get_label(delete $args->{'-fixedfontslabel'}
+				     || "Fi~xed Only");
+	$fixedcb = $bf->Checkbutton
+	  (@{$fixedfonts_res{'args'}},
+	   -variable => \$w->{Configure}{-fixedfont},
+	   -command => sub { $w->InsertFamilies; },
+	  )->grid(-column => 5, -row => 0,
+		  -sticky => 'w', -padx => 5);
+    }
+    delete $args->{'-fixedfontsbutton'};
+
     $w->grid('columnconfigure', 0, -minsize => 4);
     $w->grid('columnconfigure', 4, -minsize => 4);
     $w->grid('rowconfigure',    0, -minsize => 4);
@@ -268,21 +292,26 @@ sub Populate {
       if $altsample_res{'key'};
     $w->bind("<$nicefonts_res{'key'}>" => sub { $nicecb->invoke })
       if $nicefonts_res{'key'};
+    $w->bind("<$fixedfonts_res{'key'}>" => sub { $fixedcb->invoke })
+      if $fixedfonts_res{'key'};
 
     # XXX -subbg: ugly workaround...
     $w->ConfigSpecs
-      (-subbg       => [ 'PASSIVE', 'subBackground', 'SubBackground', 'white'],
-       -nicefont    => [ 'PASSIVE', undef, undef, 0],
-       -sampletext  => [ 'PASSIVE', undef, undef, 
-		         'The Quick Brown Fox Jumps Over The Lazy Dog.'],
-       -title       => [ 'METHOD', undef, undef, 'Choose font'],
+      (-subbg           => [ 'PASSIVE', 'subBackground', 'SubBackground',
+                             'white'],
+       -nicefont        => [ 'PASSIVE', undef, undef, 0],
+       -fixedfont       => [ 'PASSIVE', undef, undef, 0],
+       -sampletext      => [ 'PASSIVE', undef, undef, 
+		             'The Quick Brown Fox Jumps Over The Lazy Dog.'],
+       -title           => [ 'METHOD', undef, undef, 'Choose font'],
+       -customsizetitle => [ 'PASSIVE', undef, undef, 'Choose font size'],
        DEFAULT   => [ 'family_list' ],
       );
 
     $w->Delegates(DEFAULT => 'family_list');
 
     # according to the manpage, the fonts are only destroyed if the
-    # last reference to them are destroyed, too
+    # last reference to them is also destroyed
     # XXX disable for now
 #    $w->fontDelete($dialog_font, 
 #		   $bold_font, $italic_font,
@@ -343,7 +372,9 @@ sub Accept {
 }
 
 sub Show {
-    my($w, @args) = @_;
+    my($w, %args) = @_;
+
+    my $test_hack = delete $args{'-_testhack'};
 
     $w->transient($w->Parent->toplevel);
     my $oldFocus = $w->focusCurrent;
@@ -358,10 +389,11 @@ sub Show {
     $w->Subwidget('size_list')->configure(-bg => $w->cget(-subbg));
     $w->Subwidget('sample_canvas')->configure(-bg => $w->cget(-subbg));
 
-    $w->Popup(@args); 
-    $w->waitVisibility;
+    $w->Popup(%args);
+    # XXX won't work with 800.015?
+    #$w->waitVisibility;
     $w->focus;
-    $w->waitVariable(\$w->{Selected});
+    $w->waitVariable(\$w->{Selected}) unless $test_hack;
 
     eval {
 	$oldFocus->focus if $oldFocus;
@@ -401,6 +433,7 @@ sub InsertFamilies {
     eval {
 	$w->{'family_index'} = [];
 	my $nicefont = $w->cget(-nicefont); # XXX name?
+	my $fixedfont = $w->cget(-fixedfont);
 	my $curr_family = $w->fontActual($w->{'curr_font'}, -family);
 	my $famlb = $w->Subwidget('family_list');
 	$famlb->delete('all');
@@ -409,16 +442,19 @@ sub InsertFamilies {
 	my $i = 0;
 	foreach my $fam (@fam) {
 	    next if $fam eq '';
+	    next if $fixedfont
+	      and not $w->fontMetrics($w->Font(-family => $fam), '-fixed');
 	    (my $u_fam = $fam) =~ s/\b(.)/\u$1/g;
-	    $w->{'familiy_index'}[$i] = $fam;
+	    $w->{'family_index'}[$i] = $fam;
 	    my $f_style = $famlb->ItemStyle
-	      ('text', 
+	      ('text',
 	       ($nicefont ? (-font => "{$fam}") : ()),
-	       -bg => $bg,
+	       -background => $bg,
 	      );
 	    $famlb->add($i, -text => $u_fam, -style => $f_style);
 	    if ($curr_family eq $fam) {
 		$famlb->selectionSet($i);
+		$famlb->anchorSet($i);
 		$famlb->see($i);
 	    }
 	    $i++;
@@ -446,8 +482,57 @@ sub _get_label {
     %res;
 }
 
+sub _custom_size {
+    my($w) = @_;
+    my $t = $w->Toplevel;
+    my $label = $w->cget(-customsizetitle);
+    $t->title($label);
+
+    my $sizelb = $w->Subwidget("size_list");
+    my $fontsize = 10;
+    if (defined $sizelb->info("selection")) {
+	$fontsize = $sizelb->entrycget($sizelb->info("selection"), -text);
+    }
+
+    my $f1 = $t->Frame->pack;
+    $f1->Label(-text => $label)->pack(-side => 'left');
+    my $e = $f1->Entry(-width => 4,
+		       -textvariable => \$fontsize)->pack(-side => "left");
+    $e->focus;
+    $e->selectionRange(0,'end');
+    $e->icursor('end');
+
+    my $f = $t->Frame->pack;
+    my $waitvar = 0;
+    my $ok = $f->Button
+      (-text => "Ok",
+       -command => sub {
+	   $w->UpdateFont(-size => $fontsize);
+	   $sizelb->selectionClear;
+	   $sizelb->anchorClear;
+	   foreach ($sizelb->info("children")) {
+	       if ($sizelb->entrycget($_, -text) eq $fontsize) {
+		   $sizelb->selectionSet($_);
+		   $sizelb->anchorSet($_);
+		   $sizelb->see($_);
+		   last;
+	       }
+	   }
+	   $waitvar = 1;
+       })->pack(-side => "left");
+    $f->Button(-text => "Cancel",
+	       -command => sub { $waitvar = -1 })->pack(-side => "left");
+
+    $e->bind("<Return>" => sub { $ok->invoke });
+
+    $t->Popup(-popover => "cursor");
+    $t->waitVariable(\$waitvar);
+    $t->destroy;
+}
+
 # put some dirt into Tk::Widget...
-package Tk::Widget;
+package       # hide from CPAN indexer
+  Tk::Widget;
 
 # XXX Refont Canvases?
 sub RefontTree {
@@ -488,7 +573,22 @@ Tk::FontDialog - a font dialog widget for perl/Tk
 
 =head1 DESCRIPTION
 
-Tk::FontDialog implements a font dialog widget. XXX
+Tk::FontDialog implements a font dialog widget.
+
+In the Family and Size listboxes, the font family and font size can be
+specified. The checkbuttons on the right turn on bold, italic,
+underlined and overstriked variants of the chosen font. A sample of
+the font is shown in the middle area.
+
+With the "Alt sample" checkbutton, it is possible to show all
+characters in the charset instead of the default text. "Fixed only"
+restricts the font family list to fixed fonts only. If the "Nicefonts"
+checkbutton is set, then the font names in the listbox are displayed
+in the corresponding font. Note that this option can be slow if a lot
+of fonts are installed or for 16 bit fonts.
+
+A click with the right button in the font size listbox pops up a
+window to enter arbitrary font sizes.
 
 =head1 WIDGET-SPECIFIC OPTIONS
 
@@ -509,9 +609,22 @@ A list of font sizes. The default contains sizes from 0 to 72 points
 
 =item -nicefont
 
-If set, the Nicefonts button is activated. This means that the font
-names are displayed in its font style. This may be slow, especially if
-you have many fonts or 16 bit fonts (e.g. Asian fonts).
+If set, font names are displayed in its font style. This may be slow,
+especially if you have many fonts or 16 bit fonts (e.g. Asian fonts).
+
+=item -nicefontsbutton
+
+If set to false, then the "Nice fonts" button is not displayed.
+
+=item -fixedfont
+
+If set, proportional font families are not listed, leaving only the fixed
+fonts. This is slow, as each font must be checked to see if it is fixed
+or proportional.
+
+=item -fixedfontsbutton
+
+If set to false, then the "Fixed fonts" button is not displayed.
 
 =item -sampletext
 
@@ -552,12 +665,24 @@ setting:
 
 =item -nicefontslabel (Nicefonts)
 
+=item -fixedfontslabel (Fixed Only)
+
+=item -title (Choose font)
+
+=item -customsizetitle (Choose font size)
+
 =back
 
+=head1 CAVEAT
+
+Note that font names with whitespace like "New century schoolbook" or
+"MS Sans Serif" can cause problems when using in a -font option. The
+solution is to put the names in Tcl-like braces, like
+
+    -font => "{New century schoolbook} 10"
+
 =head1 BUGS/TODO
 
-  - better POD
-  - XXX
   - ConfigSpecs handling is poor
     put at least -font into configspecs
   - run test, call dialog for 2nd time: immediate change of font?
@@ -573,6 +698,8 @@ L<Tk::font|Tk::font>
 
 Slaven Rezic <eserte@cs.tu-berlin.de>
 
+Suggestions by Michael Houghton <herveus@Radix.Net>.
+
 =head1 COPYRIGHT
 
 Copyright (c) 1998,1999 Slaven Rezic. All rights reserved.
